"""Convert blocks to pdf file.

Code originally from https://github.com/lschwetlick/maxio through
https://github.com/chemag/maxio .
"""

import logging
from tempfile import NamedTemporaryFile
from subprocess import check_call

from .svg import rm_to_svg

_logger = logging.getLogger(__name__)


def rm_to_pdf(rm_path, pdf_path, debug=0):
    """Convert `rm_path` to PDF at `pdf_path`."""
    with NamedTemporaryFile(suffix=".svg") as f_temp:
        rm_to_svg(rm_path, f_temp.name)

        # use inkscape to convert svg to pdf
        check_call(["inkscape", f_temp.name, "--export-filename", pdf_path])


def svg_to_pdf(svg_file, pdf_file):
    """Read svg data from `svg_file` and write PDF data to `pdf_file`."""

    with NamedTemporaryFile("wt", suffix=".svg") as fsvg, NamedTemporaryFile("rb", suffix=".pdf") as fpdf:
        fsvg.write(svg_file.read())

        # use inkscape to convert svg to pdf
        check_call(["inkscape", fsvg.name, "--export-filename", fpdf.name])

        pdf_file.write(fpdf.read())



"""Export text content of rm files as Markdown."""

from rmscene import read_tree
from rmscene import scene_items as si

import logging



def print_text(f, fout):
    tree = read_tree(f)

    # Find out what anchor characters are used
    anchor_ids = set(collect_anchor_ids(tree.root))

    if tree.root_text:
        print_root_text(tree.root_text, fout, anchor_ids)

    JOIN_TOLERANCE = 2
    print("\n\n# Highlights", file=fout)
    last_pos = 0
    for item in tree.walk():
        if isinstance(item, si.GlyphRange):
            if item.start > last_pos + JOIN_TOLERANCE:
                print(file=fout)
            print(">", item.text, file=fout)
            last_pos = item.start + len(item.text)
    print(file=fout)


def print_root_text(root_text, fout, anchor_ids):
    for fmt, line, ids in root_text.formatted_lines_with_ids():
        annotated_line = annotate_anchor_ids(anchor_ids, line, ids)
        if fmt == si.TextFormat.BULLET:
            fout.write("- " + annotated_line)
        elif fmt == si.TextFormat.BULLET2:
            fout.write("  + " + annotated_line)
        elif fmt == si.TextFormat.BOLD:
            fout.write("> " + annotated_line)
        elif fmt == si.TextFormat.HEADING:
            fout.write("# " + annotated_line)
        elif fmt == si.TextFormat.PLAIN:
            fout.write(annotated_line)
        else:
            fout.write(("[unknown format %s] " % fmt) + annotated_line)


def annotate_anchor_ids(anchor_ids, line, ids):
    """Annotate appearances of `anchor_ids` in `line`."""
    result = ""
    for char, char_id in zip(line, ids):
        if char_id in anchor_ids:
            result += f"<<{char_id.part1},{char_id.part2}>>"
        result += char
    return result


def collect_anchor_ids(item):
    if isinstance(item, si.Group):
        if item.anchor_id is not None:
            yield item.anchor_id.value
        for child in item.children.values():
            yield from collect_anchor_ids(child)



"""CLI for converting rm files."""

import os
import sys
import io
from pathlib import Path
from contextlib import contextmanager
import click
from rmscene import read_tree, read_blocks, write_blocks, simple_text_document
from .exporters.svg import tree_to_svg
from .exporters.pdf import svg_to_pdf
from .exporters.markdown import print_text

import logging


@click.command
@click.version_option()
@click.option('-v', '--verbose', count=True)
@click.option("-f", "--from", "from_", metavar="FORMAT", help="Format to convert from (default: guess from filename)")
@click.option("-t", "--to", metavar="FORMAT", help="Format to convert to (default: guess from filename)")
@click.option("-o", "--output", type=click.Path(), help="Output filename (default: write to standard out)")
@click.argument("input", nargs=-1, type=click.Path(exists=True))
def cli(verbose, from_, to, output, input):
    """Convert to/from reMarkable v6 files.

    Available FORMATs are: `rm` (reMarkable file), `markdown`, `svg`, `pdf`,
    `blocks`, `blocks-data`.

    Formats `blocks` and `blocks-data` dump the internal structure of the `rm`
    file, with and without detailed data values respectively.

    """

    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose >= 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    input = [Path(p) for p in input]
    if output is not None:
        output = Path(output)

    if from_ is None:
        if not input:
            raise click.UsageError("Must specify input filename or --from")
        from_ = guess_format(input[0])
    if to is None:
        if output is None:
            raise click.UsageError("Must specify --output or --to")
        to = guess_format(output)

    if from_ == "rm":
        with open_output(to, output) as fout:
            for fn in input:
                convert_rm(Path(fn), to, fout)
    elif from_ == "markdown":
        text = "".join(
            Path(fn).read_text() for fn in input
        )
        with open_output(to, output) as fout:
            convert_text(text, fout)
    else:
        raise click.UsageError("source format %s not implemented yet" % from_)


@contextmanager
def open_output(to, output):
    to_binary = to in ("pdf", "rm")
    if output is None:
        # Write to stdout
        if to_binary:
            with os.fdopen(sys.stdout.fileno(), "wb", closefd=False) as f:
                yield f
        else:
            yield sys.stdout
    else:
        with open(output, "w" + ("b" if to_binary else "t")) as f:
            yield f


def guess_format(p: Path):
    # XXX could be neater
    if p.suffix == ".rm":
        return "rm"
    if p.suffix == ".svg":
        return "svg"
    elif p.suffix == ".pdf":
        return "pdf"
    elif p.suffix == ".md" or p.suffix == ".markdown":
        return "markdown"
    else:
        return "blocks"


from rmscene import scene_items as si
def tree_structure(item):
    if isinstance(item, si.Group):
        return (
            item.node_id,
            (
                item.label.value,
                item.visible.value,
                (
                    item.anchor_id.value if item.anchor_id else None,
                    item.anchor_type.value if item.anchor_type else None,
                    item.anchor_threshold.value if item.anchor_threshold else None,
                    item.anchor_origin_x.value if item.anchor_origin_x else None,
                )
            ),
            [tree_structure(child) for child in item.children.values() if child],
        )
    else:
        return item


def convert_rm(filename: Path, to, fout):
    with open(filename, "rb") as f:
        if to == "blocks":
            pprint_blocks(f, fout)
        elif to == "blocks-data":
            pprint_blocks(f, fout, data=False)
        elif to == "tree":
            # Experimental dumping of tree structure
            pprint_tree(f, fout, data=True)
        elif to == "tree-data":
            # Experimental dumping of tree structure
            pprint_tree(f, fout, data=False)
        elif to == "markdown":
            print_text(f, fout)
        elif to == "svg":
            tree = read_tree(f)
            tree_to_svg(tree, fout)
        elif to == "pdf":
            buf = io.StringIO()
            tree = read_tree(f)
            tree_to_svg(tree, buf)
            buf.seek(0)
            svg_to_pdf(buf, fout)
        else:
            raise click.UsageError("Unknown format %s" % to)


def pprint_blocks(f, fout, data=True) -> None:
    import pprint
    depth = None if data else 1
    result = read_blocks(f)
    for el in result:
        print(file=fout)
        pprint.pprint(el, depth=depth, stream=fout)


def pprint_tree(f, fout, data=True) -> None:
    tree = read_tree(f)

    import pprint
    import re

    def pprint_Line(self, object, stream, indent, allowance, context, level):
        min_x = min(p.x for p in object.points)
        min_y = min(p.y for p in object.points)
        max_x = max(p.x for p in object.points)
        max_y = max(p.y for p in object.points)
        rep = re.sub(r"points=\[.*\]",
                     f"points=[({min_x: 4.0f},{min_y: 4.0f})-({max_x: 4.0f},{max_y: 4.0f})]",
                     repr(object))
        stream.write(rep)

    pprint.PrettyPrinter._dispatch[si.Line.__repr__] = pprint_Line

    depth = None if data else 1
    pprint.pprint(tree_structure(tree.root), stream=fout)
    pprint.pprint(tree_structure(tree.root_text), depth=depth, stream=fout)



def convert_text(text, fout):
    write_blocks(fout, simple_text_document(text))


if __name__ == "__main__":
    cli()



"""Common code for writing tools.

Code originally from https://github.com/lschwetlick/maxio through
https://github.com/chemag/maxio .
"""

import logging
import math

_logger = logging.getLogger(__name__)


# color_id to RGB conversion
# 1. we use "color_id" for a unique, proprietary ID for colors,
#   (see scene_stream.py):
remarkable_palette = {
    # BLACK = 0
    0: [0, 0, 0],
    # GRAY = 1
    1: [125, 125, 125],
    # WHITE = 2
    2: [255, 255, 255],
    # https://www.color-name.com/highlighter-yellow.color
    # YELLOW = 3
    3: [251, 247, 25],
    # GREEN = 4
    4: [0, 255, 0],
    # PINK = 5
    # https://www.rapidtables.com/web/color/pink-color.html
    5: [255, 192, 203],
    # BLUE = 6
    6: [0, 0, 255],
    # RED = 7
    7: [255, 0, 0],
    # GRAY_OVERLAP = 8
    8: [125, 125, 125],
}


class Pen:
    def __init__(self, base_width, base_color_id):
        self.base_width = base_width
        self.base_color = remarkable_palette[base_color_id]
        self.segment_length = 1000
        self.base_opacity = 1
        self.name = "Basic Pen"
        # initial stroke values
        self.stroke_linecap = "round"
        self.stroke_opacity = 1
        self.stroke_width = base_width
        self.stroke_color = base_color_id

    # note that the units of the points have had their units converted
    # in scene_stream.py
    # speed = d.read_float32() * 4
    # ---> replace speed with speed / 4 [input]
    # direction = 255 * d.read_float32() / (math.pi * 2)
    # ---> replace tilt with direction_to_tilt() [input]
    @classmethod
    def direction_to_tilt(cls, direction):
        return direction * (math.pi * 2) / 255
    # width = int(round(d.read_float32() * 4))
    # ---> replace width with width / 4 [input]
    # ---> replace width with 4 * width [output]
    # pressure = d.read_float32() * 255
    # ---> replace pressure with pressure / 255 [input]

    def get_segment_width(self, speed, direction, width, pressure, last_width):
        return self.base_width

    def get_segment_color(self, speed, direction, width, pressure, last_width):
        return "rgb"+str(tuple(self.base_color))

    def get_segment_opacity(self, speed, direction, width, pressure, last_width):
        return self.base_opacity

    def cutoff(self, value):
        """must be between 1 and 0"""
        value = 1 if value > 1 else value
        value = 0 if value < 0 else value
        return value

    @classmethod
    def create(cls, pen_nr, color_id, width):
        # print(f'----> create(cls, pen_nr: {pen_nr}, color_id: {color_id}, width: {width})')
        # Brush
        if pen_nr == 0 or pen_nr == 12:
            return Brush(width, color_id)
        # caligraphy
        elif pen_nr == 21:
            return Caligraphy(width, color_id)
        # Marker
        elif pen_nr == 3 or pen_nr == 16:
            return Marker(width, color_id)
        # BallPoint
        elif pen_nr == 2 or pen_nr == 15:
            return Ballpoint(width, color_id)
        # Fineliner
        elif pen_nr == 4 or pen_nr == 17:
            return Fineliner(width, color_id)
        # Pencil
        elif pen_nr == 1 or pen_nr == 14:
            return Pencil(width, color_id)
        # Mechanical Pencil
        elif pen_nr == 7 or pen_nr == 13:
            return Mechanical_Pencil(width, color_id)
        # Highlighter
        elif pen_nr == 5 or pen_nr == 18:
            width = 15
            return Highlighter(width, color_id)
        # Erase area
        elif pen_nr == 8:
            return Erase_Area(width, color_id)
        # Eraser
        elif pen_nr == 6:
            color_id = 2
            return Eraser(width, color_id)
        raise Exception(f'Unknown pen_nr: {pen_nr}')


class Fineliner(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.base_width = (base_width ** 2.1) * 1.3
        self.name = "Fineliner"


class Ballpoint(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.segment_length = 5
        self.name = "Ballpoint"

    def get_segment_width(self, speed, direction, width, pressure, last_width):
        segment_width = (0.5 + pressure / 255) + (1 * width / 4) - 0.5*((speed / 4)/50)
        return segment_width

    def get_segment_color(self, speed, direction, width, pressure, last_width):
        intensity = (0.1 * - ((speed / 4) / 35)) + (1.2 * pressure / 255) + 0.5
        intensity = self.cutoff(intensity)
        # using segment color not opacity because the dots interfere with each other.
        # Color must be 255 rgb
        segment_color = [int(abs(intensity - 1) * 255)] * 3
        return "rgb"+str(tuple(segment_color))

    # def get_segment_opacity(self, speed, direction, width, pressure, last_width):
    #     segment_opacity = (0.2 * - ((speed / 4) / 35)) + (0.8 * pressure / 255)
    #     segment_opacity *= segment_opacity
    #     segment_opacity = self.cutoff(segment_opacity)
    #     return segment_opacity


class Marker(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.segment_length = 3
        self.name = "Marker"

    def get_segment_width(self, speed, direction, width, pressure, last_width):
        segment_width = 0.9 * ((width / 4) - 0.4 * self.direction_to_tilt(direction)) + (0.1 * last_width)
        return segment_width


class Pencil(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.segment_length = 2
        self.name = "Pencil"

    def get_segment_width(self, speed, direction, width, pressure, last_width):
        segment_width = 0.7 * ((((0.8*self.base_width) + (0.5 * pressure / 255)) * (width / 4)) - (0.25 * self.direction_to_tilt(direction)**1.8) - (0.6 * (speed / 4) / 50))
        # segment_width = 1.3*(((self.base_width * 0.4) * pressure) - 0.5 * ((self.direction_to_tilt(direction) ** 0.5)) + (0.5 * last_width))
        max_width = self.base_width * 10
        segment_width = segment_width if segment_width < max_width else max_width
        return segment_width

    def get_segment_opacity(self, speed, direction, width, pressure, last_width):
        segment_opacity = (0.1 * - ((speed / 4) / 35)) + (1 * pressure / 255)
        segment_opacity = self.cutoff(segment_opacity) - 0.1
        return segment_opacity


class Mechanical_Pencil(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.base_width = self.base_width ** 2
        self.base_opacity = 0.7
        self.name = "Mechanical Pencil"


class Brush(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.segment_length = 2
        self.stroke_linecap = "round"
        self.opacity = 1
        self.name = "Brush"

    def get_segment_width(self, speed, direction, width, pressure, last_width):
        segment_width = 0.7 * (((1 + (1.4 * pressure / 255)) * (width / 4)) - (0.5 * self.direction_to_tilt(direction)) - ((speed / 4) / 50))  # + (0.2 * last_width)
        return segment_width

    def get_segment_color(self, speed, direction, width, pressure, last_width):
        intensity = ((pressure / 255) ** 1.5 - 0.2 * ((speed / 4) / 50)) * 1.5
        intensity = self.cutoff(intensity)
        # using segment color not opacity because the dots interfere with each other.
        # Color must be 255 rgb
        rev_intensity = abs(intensity - 1)
        segment_color = [int(rev_intensity * (255 - self.base_color[0])),
                         int(rev_intensity * (255 - self.base_color[1])),
                         int(rev_intensity * (255 - self.base_color[2]))]

        return "rgb"+str(tuple(segment_color))


class Highlighter(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.stroke_linecap = "square"
        self.base_opacity = 0.3
        self.stroke_opacity = 0.2
        self.name = "Highlighter"


class Eraser(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.stroke_linecap = "square"
        self.base_width = self.base_width * 2
        self.name = "Eraser"


class Erase_Area(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.stroke_linecap = "square"
        self.base_opacity = 0
        self.name = "Erase Area"


class Caligraphy(Pen):
    def __init__(self, base_width, base_color_id):
        super().__init__(base_width, base_color_id)
        self.segment_length = 2
        self.name = "Calligraphy"

    def get_segment_width(self, speed, direction, width, pressure, last_width):
        segment_width = 0.9 * (((1 + pressure / 255) * (width / 4)) - 0.3 * self.direction_to_tilt(direction)) + (0.1 * last_width)
        return segment_width



import subprocess
import sys


def run_command(command, **kwargs):
    env = kwargs.get('env', None)
    stdin = subprocess.PIPE if kwargs.get('stdin', False) else None
    bufsize = kwargs.get('bufsize', 0)
    universal_newlines = kwargs.get('universal_newlines', False)
    default_close_fds = True if sys.platform == 'linux2' else False
    close_fds = kwargs.get('close_fds', default_close_fds)
    shell = type(command) in (type(''), type(u''))
    dry_run = kwargs.get('dry_run', False)
    if dry_run:
        return 0, b'stdout', b'stderr'
    p = subprocess.Popen(command, stdin=stdin, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, bufsize=bufsize,
                         universal_newlines=universal_newlines,
                         env=env, close_fds=close_fds, shell=shell)
    # wait for the command to terminate
    if stdin is not None:
        out, err = p.communicate(stdin)
    else:
        out, err = p.communicate()
    returncode = p.returncode
    # clean up
    del p
    # return results
    return returncode, out, err



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rmc', 'rmc.exporters']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0,<9.0', 'rmscene>=0.3.0,<0.4.0']

entry_points = \
{'console_scripts': ['rmc = rmc.cli:cli']}

setup_kwargs = {
    'name': 'rmc',
    'version': '0.2.0a0',
    'description': 'Convert to/from v6 .rm files from the reMarkable tablet',
    'long_description': 'None',
    'author': 'Rick Lupton',
    'author_email': 'mail@ricklupton.name',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Convert blocks to svg file.

Code originally from https://github.com/lschwetlick/maxio through
https://github.com/chemag/maxio .
"""

import logging
import string
from pathlib import Path

from typing import Iterable

from rmscene import scene_items as si
from rmscene import (
    read_tree,
    SceneTree,
    CrdtId,
)

from .writing_tools import (
    Pen,
)

_logger = logging.getLogger(__name__)


SCREEN_WIDTH = 1404
SCREEN_HEIGHT = 1872
SCREEN_DPI = 226

SCALE = 72.0 / SCREEN_DPI

PAGE_WIDTH_PT = SCREEN_WIDTH * SCALE
PAGE_HEIGHT_PT = SCREEN_HEIGHT * SCALE
X_SHIFT = PAGE_WIDTH_PT // 2


def xx(screen_x):
    return screen_x * SCALE #+ X_SHIFT


def yy(screen_y):
    return screen_y * SCALE


TEXT_TOP_Y = -88
LINE_HEIGHTS = {
    # Tuned this line height using template grid -- it definitely seems to be
    # 71, rather than 70 or 72. Note however that it does interact a bit with
    # the initial text y-coordinate below.
    si.TextFormat.PLAIN: 71,
    si.TextFormat.BULLET: 35,
    si.TextFormat.BOLD: 70,
    si.TextFormat.HEADING: 150,

    # There appears to be another format code (value 0) which is used when the
    # text starts far down the page, which case it has a negative offset (line
    # height) of about -20?
    #
    # Probably, actually, the line height should be added *after* the first
    # line, but there is still something a bit odd going on here.
}


# <html>
# <body>
# <div style="border: 1px solid grey; margin: 2em; float: left;">
# <svg xmlns="http://www.w3.org/2000/svg" height="$height" width="$width">
SVG_HEADER = string.Template("""
<svg xmlns="http://www.w3.org/2000/svg" height="$height" width="$width" viewbox="$viewbox">
    <script type="application/ecmascript"> <![CDATA[
        var visiblePage = 'p1';
        function goToPage(page) {
            document.getElementById(visiblePage).setAttribute('style', 'display: none');
            document.getElementById(page).setAttribute('style', 'display: inline');
            visiblePage = page;
        }
    ]]>
    </script>
""")


def rm_to_svg(rm_path, svg_path):
    """Convert `rm_path` to SVG at `svg_path`."""
    with open(rm_path, "rb") as infile, open(svg_path, "wt") as outfile:
        tree = read_tree(infile)
        tree_to_svg(tree, outfile)


def read_template_svg(template_path: Path) -> str:
    lines = template_path.read_text().splitlines()
    return "\n".join(lines[2:-1])


def tree_to_svg(tree: SceneTree, output, include_template=None):
    """Convert Blocks to SVG."""

    # add svg header
    # output.write('<svg xmlns="http://www.w3.org/2000/svg">\n')
    output.write(SVG_HEADER.substitute(width=PAGE_WIDTH_PT,
                                       height=PAGE_HEIGHT_PT,
                                       viewbox=f"0 0 {PAGE_WIDTH_PT} {PAGE_HEIGHT_PT}") + "\n")

    if include_template is not None:
        template = read_template_svg(include_template)
        output.write(f'    <g id="template" style="display:inline" transform="scale({SCALE})">\n')
        output.write(template)
        output.write('    </g>\n')

    output.write(f'    <g id="p1" style="display:inline" transform="translate({X_SHIFT},0)">\n')
    output.write('        <filter id="blurMe"><feGaussianBlur in="SourceGraphic" stdDeviation="10" /></filter>\n')

    # These special anchor IDs are for the top and bottom of the page.
    anchor_pos = {
        CrdtId(0, 281474976710654): 270,
        CrdtId(0, 281474976710655): 700,
    }
    if tree.root_text is not None:
        draw_text(tree.root_text, output, anchor_pos)
    _logger.debug("anchor_pos: %s", anchor_pos)

    draw_group(tree.root, output, anchor_pos)

    # # Overlay the page with a clickable rect to flip pages
    # output.write('\n')
    # output.write('        <!-- clickable rect to flip pages -->\n')
    # output.write(f'        <rect x="0" y="0" width="{svg_doc_info.width}" height="{svg_doc_info.height}" fill-opacity="0"/>\n')
    # Closing page group
    output.write('    </g>\n')
    # END notebook
    output.write('</svg>\n')


def draw_group(item: si.Group, output, anchor_pos):
    anchor_x = 0.0
    anchor_y = 0.0
    if item.anchor_id is not None:
        assert item.anchor_origin_x is not None
        anchor_x = item.anchor_origin_x.value
        if item.anchor_id.value in anchor_pos:
            anchor_y = anchor_pos[item.anchor_id.value]
            _logger.debug("Group anchor: %s -> y=%.1f", item.anchor_id.value, anchor_y)
        else:
            _logger.warning("Group anchor: %s is unknown!", item.anchor_id.value)
    output.write(f'    <g id="{item.node_id}" transform="translate({xx(anchor_x)}, {yy(anchor_y)})">\n')
    for child_id in item.children:
        child = item.children[child_id]
        _logger.debug("Group child: %s %s", child_id, type(child))
        output.write(f'    <!-- child {child_id} -->\n')
        if isinstance(child, si.Group):
            draw_group(child, output, anchor_pos)
        elif isinstance(child, si.Line):
            draw_stroke(child, output)
    output.write(f'    </g>\n')


def draw_stroke(item: si.Line, output):
    _logger.debug("Writing line: %s", item)

    # initiate the pen
    pen = Pen.create(item.tool.value, item.color.value, item.thickness_scale/10)
    K = 5

    # BEGIN stroke
    output.write(f'        <!-- Stroke tool: {item.tool.name} color: {item.color.name} thickness_scale: {item.thickness_scale} -->\n')
    output.write('        <polyline ')
    output.write(f'style="fill:none;stroke:{pen.stroke_color};stroke-width:{pen.stroke_width/K};opacity:{pen.stroke_opacity}" ')
    output.write(f'stroke-linecap="{pen.stroke_linecap}" ')
    output.write('points="')

    last_xpos = -1.
    last_ypos = -1.
    last_segment_width = segment_width = 0
    # Iterate through the point to form a polyline
    for point_id, point in enumerate(item.points):
        # align the original position
        xpos = point.x
        ypos = point.y
        if point_id % pen.segment_length == 0:
            segment_color = pen.get_segment_color(point.speed, point.direction, point.width, point.pressure, last_segment_width)
            segment_width = pen.get_segment_width(point.speed, point.direction, point.width, point.pressure, last_segment_width)
            segment_opacity = pen.get_segment_opacity(point.speed, point.direction, point.width, point.pressure, last_segment_width)
            # print(segment_color, segment_width, segment_opacity, pen.stroke_linecap)
            # UPDATE stroke
            output.write('"/>\n')
            output.write('        <polyline ')
            output.write(f'style="fill:none; stroke:{segment_color} ;stroke-width:{segment_width/K:.3f};opacity:{segment_opacity}" ')
            output.write(f'stroke-linecap="{pen.stroke_linecap}" ')
            output.write('points="')
            if last_xpos != -1.:
                # Join to previous segment
                output.write(f'{xx(last_xpos):.3f},{yy(last_ypos):.3f} ')
        # store the last position
        last_xpos = xpos
        last_ypos = ypos
        last_segment_width = segment_width

        # BEGIN and END polyline segment
        output.write(f'{xx(xpos):.3f},{yy(ypos):.3f} ')

    # END stroke
    output.write('" />\n')


def draw_text(text: si.Text, output, anchor_pos):
    output.write('    <g class="root-text" style="display:inline">\n')

    # add some style to get readable text
    output.write('''
    <style>
        text.heading {
            font: 14pt serif;
        }
        text.bold {
            font: 8pt sans-serif bold;
        }
        text, text.plain {
            font: 7pt sans-serif;
        }
    </style>
    ''')

    y_offset = TEXT_TOP_Y
    for fmt, line, ids in text.formatted_lines_with_ids():
        y_offset += LINE_HEIGHTS[fmt]

        xpos = text.pos_x
        ypos = text.pos_y + y_offset
        cls = fmt.name.lower()
        if line:
            output.write(f'        <!-- Text line char_id: {ids[0]} -->\n')
            output.write(f'        <text x="{xx(xpos)}" y="{yy(ypos)}" class="{cls}">{line.strip()}</text>\n')

        # Save y-coordinates of potential anchors
        for k in ids:
            anchor_pos[k] = ypos

    output.write('    </g>\n')
